################################################################################
# File    : ast2cpp.tcl
# Content : cpp code generation used by hsm2cpp.tcl
# History : 09-Feb-2004 - [rb] splitted hsm2cpp.tcl into parser.tcl,
#                              hsm2ast.tcl, ast2cpp.tcl & hsm2cpp.tcl
################################################################################

#   Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



# N.B. #########################################################################
# The functions here are using the globals : index


# code generation ##############################################################

# genHsm --
#
#   generate Hsm code
#
# return
#   {header code}

proc genHsm {hsm {level 0} {super "__NONE__"}} {
	global index
	
	if {[lindex $hsm end] != "definition"} {
		return [list "" ""]
	}
	
	set name      [lindex $hsm $index(name)]
	set init      [lindex $hsm $index(init)]
	set enter     [lindex $hsm $index(enter)]
	set leave     [lindex $hsm $index(leave)]
	set timer     [lindex $hsm $index(timer)]
	set actions   [lindex $hsm $index(actions)]
	set subHsms   [lindex $hsm $index(subHsms)]
	set outInputs [lindex $hsm $index(outInputs)]
	set intInputs [lindex $hsm $index(intInputs)]
	set outVars   [lindex $hsm $index(outVars)]
	set intVars   [lindex $hsm $index(intVars)]
	set initials  [lindex $hsm $index(initials)]
	set methods   [lindex $hsm $index(methods)]
	set requis    [lindex $hsm $index(requis)]
	set virtuals  [lindex $hsm $index(virtuals)]
	set locals    [lindex $hsm $index(locals)]
	set hsmInputs [lindex $hsm $index(hsmInputs)]

	set first [lindex [lindex $subHsms $index(name)] 0]
	set ind [string repeat "\t" $level]

	### begin of the namespace dedicated to the current state ###
	set decl "\n/* $name hsm definition [string repeat "*" [expr 59 - [string length $name]]]*/\n\n"
	set impl "\n/* $name hsm implementation [string repeat "*" [expr 55 - [string length $name]]]*/\n\n"

	append decl "${ind}namespace $name \{\n\n"
	append impl "${ind}namespace $name \{\n\n"

	### events interface ###
	if {$super == "__NONE__"} {
		set methods [genMethods $methods]
		append decl [lindex $methods 0]
		append impl [lindex $methods 1]
		set super "Events"
	}
	

	### Data definition ###
	append decl "${ind}\t// Data\n"
	append decl [genData $ind $outVars $intVars $outInputs $intInputs]
	

	### AHsm definition ###
	append decl "${ind}\t// Abstract Hsm\n"
	append decl "${ind}\tclass AHsm : public hsm::Hsm, protected virtual Data \{\n"

	append decl "${ind}\tprotected :\n"

	### init
	if {($init != "") || ($subHsms != {})} {
		set init [genInit $ind $subHsms $init]
		append decl [lindex $init 0]
		append impl [lindex $init 1]
	}

	### checkRequis
	if {($requis != {})} {
		set check [genCheckRequis $ind $requis]
		append decl [lindex $check 0]
		append impl [lindex $check 1]	
	}
	
	### enter, leave, escape
	if {($enter != "") || ($timer != "")} {
		set enter [genEnterLeave $ind enter $enter $timer]
		append decl [lindex $enter 0]
		append impl [lindex $enter 1]
	}

	if {$leave != ""} {
		set leave [genEnterLeave $ind leave $leave]
		append decl [lindex $leave 0]
		append impl [lindex $leave 1]
	}

	### input registration
	if {$hsmInputs != {}} {
		set register [genRegister $ind $hsmInputs]
		append decl [lindex $register 0]
		append impl [lindex $register 1]
	}

	### initial state
	if {$initials != {}} {
		set getInitial [genGetInitial $ind $initials]
		append decl [lindex $getInitial 0]
		append impl [lindex $getInitial 1]
	}

	### event handling
	if {$actions != {}} {
		set handle [genHandle $ind $actions]
		append decl [lindex $handle 0]
		append impl [lindex $handle 1]
	}
	
	### local methods
	if {$locals != {}} {
		append decl "\n[genLocalMethods $ind $locals]"
	}
	
	### on the fly input changing
	if {$outInputs != {}} {
		append decl "${ind}\tpublic :\n"
		set inputSetters [genInputSetters $ind $outInputs]
		append decl [lindex $inputSetters 0]
		append impl [lindex $inputSetters 1]
	}

	append decl "${ind}\t\};\n\n"


	### Interface definition ###
	append decl "${ind}\t// Interface\n"
	append decl [genInterface $ind $super $virtuals]


	### Hsm typedef
	append decl "${ind}\t// Hsm\n"
	append decl "${ind}\textern const char name\[\];\n"
	append impl "${ind}\tconst char name\[\] = \"[string tolower $name 0 0]\";\n\n"
	append decl "${ind}\ttemplate< class SuperInterface = ${super}::Interface >\n"
	append decl "${ind}\tstruct T {\n"
	append decl "${ind}\t\ttypedef hsm::impl::Hsm< name, AHsm, TInterface< SuperInterface >, Data > Hsm;\n"
	append decl "${ind}\t};\n"
	append decl "${ind}\ttypedef T< >::Hsm Hsm;\n\n"


	### substates ###
	foreach subHsm $subHsms {
		set sub [genHsm $subHsm [expr $level + 1] $name]
		append decl [lindex $sub 0]
		append impl [lindex $sub 1]
	}

	append decl "${ind}\} // namespace $name\n\n"
	append impl "${ind}\} // namespace $name\n\n"

	return [list $decl $impl]
}


# genData --
#
#   Generate Data structure
#
# parameters
#   ind       indentation string
#   outVars   list of undefined variables
#   intVars   list of defined variables
#   outInputs list of undefined inputs
#   intInputs list of defined inputs
#
# result
#   header

proc genData {ind outVars intVars outInputs intInputs} {
	append decl "${ind}\tstruct Data \{\n"

	set ctorArgs {}
	set ctorInits {}
	set ctorDInits {}

	### state variables
	foreach var $outVars {
		set varType [lindex $var 0]
		set varName [lindex $var 1]
		switch [string range $varType end end] {
			"&" {
				append decl "${ind}\t\t$varType $varName;\n"
				lappend ctorArgs "$varType $varName"
				lappend ctorInits "${varName}($varName)"
				lappend ctorDInits "${varName}(*reinterpret_cast< [string range $varType 0 end-1] * >(0))"			
			}
			default {
				append decl "${ind}\t\t$varType &$varName;\n"
				lappend ctorArgs "$varType &$varName"
				lappend ctorInits "${varName}($varName)"
				lappend ctorDInits "${varName}(*reinterpret_cast< $varType * >(0))"			
			}
		}
	}
	foreach var $intVars {
		set varType [lindex $var 0]
		set varName [lindex $var 1]
		set varVal  [lindex $var 2]
		if {($varVal == "__EMPTY__")} {
			set varVal ""
		}
		switch [string range $varType end end] {
			"&" {
				append decl "${ind}\t\t$varType $varName;\n"
				lappend ctorInits "${varName}($varVal)"
				lappend ctorDInits "${varName}(*reinterpret_cast< [string range $varType 0 end-1] * >(0))"
			}
			default {
				append decl "${ind}\t\t$varType &$varName;\n"
				lappend ctorInits "${varName}(*new ${varType}($varVal))"
				lappend ctorDInits "${varName}(*reinterpret_cast< $varType * >(0))"
			}
		}
	}
	
	set ctorArgs2  $ctorArgs
	set ctorInits2 $ctorInits
	
	### inputs
	foreach input $outInputs {
		set inputName [lindex $input 1]
		append decl "${ind}\t\thsm::Component *&$inputName;\n"
		lappend ctorArgs   "hsm::Component *&$inputName"
		lappend ctorInits  "${inputName}($inputName)"
		lappend ctorArgs2  "hsm::Component &$inputName"
		lappend ctorInits2 "${inputName}(*new hsm::Component *(&$inputName))"
		lappend ctorDInits "${inputName}(*reinterpret_cast< hsm::Component ** >(0))"
	}
	foreach input $intInputs {
		set inputName [lindex $input 1]
		set inputVal  [lindex $input 2]
		append decl "${ind}\t\thsm::Component *&$inputName;\n"
		lappend ctorInits  "${inputName}(*new hsm::Component *($inputVal))"
		lappend ctorInits2 "${inputName}(*new hsm::Component *($inputVal))"
		lappend ctorDInits "${inputName}(*reinterpret_cast< hsm::Component ** >(0))"
	}

	### constructors
	append decl "${ind}\t\tData([join $ctorArgs ", "])"
	if { $ctorInits != {} } {
		append decl " : [join $ctorInits ", "]"
	}
	append decl " \{\}\n"

	if { $outInputs != {} } {
		append decl "${ind}\t\tData([join $ctorArgs2 ", "])"
		if { $ctorInits2 != {} } {
			append decl " : [join $ctorInits2 ", "]"
		}
		append decl " \{\}\n"
	}
	
	if { $ctorArgs != {} } {
		append decl "${ind}\t\tData()"
		if { $ctorDInits != {} } {
			append decl " : [join $ctorDInits ", "]"
		}
		append decl " \{\}\n"
	}

	append decl "${ind}\t\};\n\n"

	return $decl
}




# genInterface --
#
#   generate Interface structure
#
# parameters
#   ind      indentation string
#   super    name of the super state
#   virtuals list of the virtual methods
#
# return
#   header

proc genInterface {ind super virtuals} {
	append decl "${ind}\ttemplate< class SuperInterface = ${super}::Interface >\n"
	append decl "${ind}\tstruct TInterface : public SuperInterface, protected virtual Data \{\n"
	foreach virtual $virtuals {
		set name [string trimleft [lindex $virtual 0]]
		set code [split [lindex $virtual 1] "\n"]
		append decl "${ind}\t\tvirtual $name \{\n"
		foreach line $code {
			set l [string trim $line]
			if {$l == ""} {} else { append decl "${ind}\t\t\t$l\n" }
		}
		append decl "${ind}\t\t\}\n"
	}
	append decl "${ind}\t\};\n"
	append decl "${ind}\ttypedef TInterface< > Interface;\n\n"

	return $decl
}


# genInit --
#
#   Generate Hsm initialisation code
#
# parameters
#   ind     indentation string
#   subHsms substates
#   code    init code
#
# return
#  {header body}

proc genInit {ind subHsms code} {
	global index

	append decl "${ind}\t\tvirtual void _init_();\n"
	append impl "${ind}\tvoid AHsm::_init_() \{\n"
	set ini "_initial_ = "
	foreach subHsm $subHsms {
		set subName [lindex $subHsm $index(name)]
		append impl "${ind}\t\t${ini}hsmMap\[${subName}::name\] = new ${subName}::T< Interface >::Hsm(this"

		set subCtorArgs {}
		set subOutInputs [lindex $subHsm $index(outInputs)]
		set subOutVars   [lindex $subHsm $index(outVars)]
		foreach var $subOutVars {
			lappend subCtorArgs [lindex $var 1]
		}
		foreach input $subOutInputs {
			lappend subCtorArgs [lindex $input 1]
		}

		if {$subCtorArgs != {}} {
			append impl ", ${subName}::Data([join $subCtorArgs ", "])"
		}

		append impl ");\n"
		set ini ""
	}

	foreach line [split $code "\n"] {
		regsub -all -- {^[ \t]*} $line {} l
		if {$l == ""} {} else { append impl "${ind}\t\t$l\n" }
	}

	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}


# genCheckRequis --
#
#   Generate checkRequis code
#
# parameters
#   ind      indentation string
#   requis   requirements definition
#
# return
#   {header body}

proc genCheckRequis {ind requis} {
	append decl "${ind}\t\tvirtual hsm::Hsm *checkRequis();\n"
	append impl "${ind}\thsm::Hsm *AHsm::checkRequis() \{\n"
	append impl "${ind}\t\thsm::Hsm *next = 0;\n"
	foreach requi $requis {
		set condition  [lindex $requi 0]
		set nextHsm    [lindex $requi 1]
		set nextValues [lrange $requi 2 end-1]
		set code       [lindex $requi end]
		append impl "${ind}\t\tif((next == 0) && !($condition)) \{\n"
		foreach line [split $code "\n"] {
			regsub -all -- {^[ \t]*} $line {} l
			if {$l == ""} {} else { append impl "${ind}\t\t\t$l\n" }
		}

		append impl "${ind}\t\t\tnext = _super_->resolve(${nextHsm}::name);\n"
		if {$nextValues != {}} {
			append impl "${ind}\t\t\t${nextHsm}::Hsm *nextHsm = dynamic_cast< ${nextHsm}::Hsm * >(next);\n"
			append impl "${ind}\t\t\tif(nextHsm != 0) \{\n"
			foreach value $nextValues {
				append impl "${ind}\t\t\t\tnextHsm->data.[lindex $value 0] = [lindex $value 1];\n"
			}
			append impl "${ind}\t\t\t\}\n"
		}
		append impl "${ind}\t\t\}\n"
	}

	append impl "${ind}\t\treturn next;\n"
	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}


# genEnterLeave --
#
#   Generate enter / leave code
#
# parameters
#   ind    indentation string
#   action name of the action
#   code   code
#   timer  time out definition
#
# return
#   {header body}

proc genEnterLeave {ind action code {timer ""}} {
	append decl "${ind}\t\tvirtual void _${action}_();\n"
	append impl "${ind}\tvoid AHsm::_${action}_() \{\n"

	foreach line [split $code "\n"] {
		regsub -all -- {^[ \t]*} $line {} l
		if {$l == ""} {} else { append impl "${ind}\t\t$l\n" }
	}
	if {$timer != ""} {
		append impl "${ind}\t\ttimerDelay = $timer;\n"
	}
	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}



# genRegister --
#
#   Generate registration code
#
# parameters
#   ind       indentation string
#   hsmInputs list of hsm inputs
#
# return
#   {header body}

proc genRegister {ind hsmInputs} {
	append decl "${ind}\t\tvirtual void registerInputs();\n"
	append impl "${ind}\tvoid AHsm::registerInputs() \{\n"
	foreach input $hsmInputs {
		append impl "${ind}\t\t$input->addReceiver(top());\n"
	}
	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}



# genGetInitial --
#
#   Generate
#
# parameters
#   ind      indentation string
#   initials list of initial conditions
#
# return
#   {header body}

proc genGetInitial {ind initials} {
	append decl "${ind}\t\tvirtual hsm::Hsm *getInitial();\n"
	append impl "${ind}\thsm::Hsm *AHsm::getInitial() \{\n"
	append impl "${ind}\t\thsm::Hsm *next = 0;\n"
	foreach initial $initials {
		set condition  [lindex $initial 0]
		set nextHsm    [lindex $initial 1]
		append impl "${ind}\t\tif((next == 0) && ($condition)) \{\n"
		append impl "${ind}\t\t\tnext = hsmMap\[${nextHsm}::name\];\n"
		append impl "${ind}\t\t\}\n"
	}
	append impl "${ind}\t\tif(next == 0) next = hsm::Hsm::getInitial();\n"
	append impl "${ind}\t\treturn next;\n"
	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}


# genHandle --
#
#   Generate event handling code
#
# parameters
#   ind     indentation string
#   actions transition definitions
#
# return
#   {header body}

proc genHandle {ind actions} {
	set starActions {}
	
	append decl "${ind}\t\tvirtual bool handleEvent(const hsm::Event &event);\n"
	append impl "${ind}\tbool AHsm::handleEvent(const hsm::Event &event) \{\n"

	append impl "${ind}\t\tswitch(event.message) \{\n"

	set mes ""
	while {$actions != {}} {
		set action [lindex $actions 0]
		set actions [lrange $actions 1 end]

		set mes [lindex $action 0]
		if {$mes == "*"} {
			lappend starActions $action
		} else {
			append impl "${ind}\t\tcase hsm::Event::$mes :\n"
			while {1} {
				append impl [genHandleOneAction "$ind\t" $action]
				if {[lindex [lindex $actions 0] 0] != $mes} { break }
				set action [lindex $actions 0]
				set actions [lrange $actions 1 end]
			}
			append impl "${ind}\t\t\tbreak;\n"
		}
	}
	
	append impl "${ind}\t\t\}\n"

	# * actions 
	foreach action $starActions {
		append impl [genHandleOneAction $ind $action]
	}

	append impl "${ind}\t\treturn false;\n"
	append impl "${ind}\t\}\n\n"

	return [list $decl $impl]
}


# genHandleOneAction --
#
#   Generate code for one action
#
# parameters
#   
#   ind     indentation string
#   action  action definition
#
# return
#   implementation

proc genHandleOneAction {ind action} {
	append impl "${ind}\t\tif(true"
	set msg    [lindex $action 0]
	set sender [lindex $action 1]
	set cond   [lindex $action 2]
	set code   [lindex $action 3]
	set next   [lindex $action 4]
	set transEvent [lindex $next 0]
	set history    [lindex $next 1]
	set nextHsm    [lindex $next 2]
	set nextValues [lrange $next 3 end]
#	if {$msg != "*"} { append impl " && (event.message == hsm::Event::$msg)" }
	if {$sender != {}} { append impl " && (event.sender == $sender)" }
	if {$cond != {}} { append impl " && $cond" }
	append impl ") \{\n"
	foreach line [split $code "\n"] {
		set l [string trim $line]
		if {$l == ""} {} else { append impl "${ind}\t\t\t$l\n" }
	}
	if {$nextHsm != ""} {
		append impl "${ind}\t\t\tleaveSubHsms();\n"
		append impl "${ind}\t\t\thsm::Hsm *next = _super_->resolve(${nextHsm}::name);\n"
				
		if {$nextValues != {}} {
			append impl "${ind}\t\t\t${nextHsm}::Hsm *nextHsm = dynamic_cast< ${nextHsm}::Hsm * >(next);\n"
			append impl "${ind}\t\t\tif(nextHsm != 0) \{\n"
			foreach value $nextValues {
				append impl "${ind}\t\t\t\tnextHsm->data.[lindex $value 0] = [lindex $value 1];\n"
			}
			append impl "${ind}\t\t\t\}\n"
		}

		append impl "${ind}\t\t\tif(next != 0) \{\n"
		append impl "${ind}\t\t\t\tnext->enterSubHsms("
		if {$history == "h"} {
			append impl "true"
		}
		append impl ");\n"
		append impl "${ind}\t\t\t\}\n"

		if {$transEvent == "="} {
			append impl "${ind}\t\t\t\tpost(event);\n"
		}
	}
	append impl "${ind}\t\t\treturn true;\n"
	append impl "${ind}\t\t\}\n"

	return $impl
}


# genLocalMethods --
#
#   Generate local methods
#
# parameters
#   ind     indentation string
#   actions method definitions
#
# return
#   header

proc genLocalMethods {ind locals} {
	foreach local $locals {
		append decl "${ind}\t\t[string trimleft [lindex $local 0]] \{\n"
		foreach line [split [lindex $local 1] "\n"] {
			set l [string trim $line]
			if {$l == ""} {} else { append decl "${ind}\t\t\t$l\n" }
		}
		append decl "${ind}\t\t\}\n\n"
	}

	return $decl
}


# genInputSetters --
#
#   Generate input setters code
#
# parameters
#   ind       indentation string
#   hsmInputs list of hsm inputs
#
# return
#   {header body}

proc genInputSetters {ind hsmInputs} {
	foreach input $hsmInputs {
		set inputName [lindex $input 1]
		append decl "${ind}\t\tvoid set[string toupper $inputName 0 0](hsm::Component &c);\n"
		append impl "${ind}\tvoid AHsm::set[string toupper $inputName 0 0](hsm::Component &c) \{\n"
		append impl "${ind}\t\t$inputName->removeReceiver(top());\n"
		append impl "${ind}\t\t$inputName = &c;\n"
		append impl "${ind}\t\t$inputName->addReceiver(top());\n"
		append impl "${ind}\t\}\n\n"
	}

	return [list $decl $impl]
}


# genMethods --
#
#   Generate methods to trigger transitions
#
# parameter
#   event names list
#
# return
#   {header body}

proc genMethods {methods} {
	set impl ""
	append decl "\t// Super interface\n"
	append decl "\tnamespace Events \{\n"
	append decl "\t\tstruct Interface : public hsm::Interface \{\n"
	append decl "\t\t\tenum Id \{\n"
	set enum {"\t\t\t\tTimeOutEvent = 0"}
	foreach method $methods {
		set name [lindex $method 0]
		lappend enum "\t\t\t\t[string toupper $name 0 0]Event"
	}
	append decl "[join $enum ",\n"]\n"
	append decl "\t\t\t\};\n"
	if {$methods != {}} {
		append impl "\tnamespace Events \{\n"
		foreach method $methods {
			set name [lindex $method 0]
			set args [lindex $method 1]
			
			set argsList {}
			set argsName {}
			foreach arg $args {
				lappend argsList "[lindex $arg 0] [lindex $arg 1]"
				lappend argsName [lindex $arg 1]
			}
			
			# structure to hold function arguments
			if {$argsList != {}} {
				append decl "\n\t\t\tstruct [string toupper $name 0 0]Args \{\n"
				foreach arg $argsList {
					append decl "\t\t\t\t$arg;\n"
				}
				append decl "\t\t\t\t[string toupper $name 0 0]Args([join $argsList ", "]) "
				set sep ":"
				foreach arg $argsName {
					append decl "$sep ${arg}($arg)"
					set sep ","
				}
				append decl " {}\n"
				append decl "\t\t\t\};\n"				
			}
			
			# function definition & declaration
			append decl "\t\t\tvoid ${name}([join $argsList ", "]);\n"
			if {$argsList != {}} {
				append impl "\t\tvoid Interface::${name}([join $argsList ", "]) \{\n"
				append impl "\t\t\t[string toupper $name 0 0]Args *args = new [string toupper $name 0 0]Args([join $argsName ", "]);\n"
				append impl "\t\t\tpost(hsm::Event(this, hsm::Event::INTERNAL, [string toupper $name 0 0]Event, args));\n"
				append impl "\t\t\}\n"
			} else {
				append impl "\t\tvoid Interface::${name}() \{ post(hsm::Event(this, hsm::Event::INTERNAL, [string toupper $name 0 0]Event)); \}\n"
			}
		}
		append impl "\t\} // namespace Events\n\n\n"
	}
	append decl "\t\t\};\n"
	append decl "\t\} // namespace Events\n\n\n"

	return [list $decl $impl]
}
