################################################################################
# File    : parser.tcl
# Content : text parsing used in hsm2ast.tcl
# History : 09-Feb-2004 - [rb] splitted hsm2cpp.tcl into parser.tcl,
#                              hsm2ast.tcl, ast2cpp.tcl & hsm2cpp.tcl
################################################################################

#   Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



# N.B. #########################################################################
# The functions here are using the globals : input, parsed


# utilities ####################################################################

# exitError --
#
#   Exit after a parsing error.
#

proc exitError {mes} {
	global parsed

	puts stderr $parsed
	set lines [split $mes "\n"]
	puts stderr ">>> error : [lindex $lines 0]"
	foreach line [lrange $lines 1 end] {
		puts stderr ">>> -> $line"
	}

	exit 1
}


# warning --
#
#   Print a warning on the console
#

proc warning {mes} {
	puts stderr ">>> warning : $mes"
}


# lastWord --
#
#   extract the last word of a string
#
# return
#   the last word

proc lastWord {s} {
	set s [string trim $s]
	return [string range $s [string wordstart $s end] end]
}


# firstWords --
#
#   extract everything but the last word of a string
#
# return
#   everything but the last word

proc firstWords {s} {
	set s [string trim $s]
	return [string trim [string range $s 0 [expr [string wordstart $s end] - 1]]]
}



# token extraction #############################################################

# putBackWord --
#
#   Puts a word back into input
#

proc putBackWord {} {
	global input
	global parsed

	set e [string wordstart $parsed end]
	set word [string range $parsed $e end]
	set parsed [string range $parsed 0 [expr $e - 1]]

	set input ${word}${input}
}


# getWord --
#
#   Extract the first word of the input
#
# return
#   the first word of the input
# side effect
#   eats this word in the global 'input' appends it to 'parsed'

proc getWord {} {
	global input
	global parsed

	if {$input == ""} {
		return "__EOF__"
	}

	set e [string wordend $input 0]
	set word [string range $input 0 [expr $e - 1]]
	set input [string range $input $e end]

	append parsed $word

	return $word
}


# putBackLine --
#
#   Puts a line back into input
#

proc putBackLine {} {
	global input
	global parsed

	set lines [split $parsed "\n"]
	set line [lindex $lines end-1]
	set parsed [join [lrange $lines 0 end-2] "\n"]

	set input "${line}\n${input}"
}


# getLine --
#
#   Extract the first line of the input
#
# return
#   the first line of the input
# side effect
#   see getWord

proc getLine {} {
	global input
	global parsed
	
	if {$input == ""} {
		return "__EOF__"
	}

	set lines [split $input "\n"]
	set line [lindex $lines 0]
	set input [join [lrange $lines 1 end] "\n"]
	
	append parsed "$line\n"

	return $line
}


# getTokenExpected --
#
#   Extract the first token (non empty word) of the input and check if it is
#   in the 'tokenExpected' list (if not empty). If not, it puts back the token
#   into input stream.
#
# return
#   the next token or __UNEXPECTED__
# side effect
#   cf getWord

proc getTokenExpected { {tokenExpected {}} } {
	while {1} {
		set token [getWord]
		if {[regexp -- {[ \t\n]+} $token] == 0} { break }
	}

	if {($tokenExpected == {}) || ([lsearch -exact $tokenExpected $token] >= 0)} {
		return $token
	} else {
		putBackWord
		return "__UNEXPECTED__"
	}
}


# getToken --
#
#   Extract the first token (non empty word) of the input and check if it is
#   in the 'tokenExpected' list (if not empty)
#
# return
#   the next token or exit with error if the token is not expected
# side effect
#   cf getTokenExpected

proc getToken { {tokenExpected {}} } {
	set token [getTokenExpected $tokenExpected]
	if {$token == "__UNEXPECTED__"} {
		set token [getToken]
		set msg "unexpected  '$token\' (need '[join $tokenExpected "' or '"]')."
		if {[lsearch -exact {"." "-" "<"} $token] >= 0} {
			append msg "\nyou may have to parentize expression containig '.' '->' or '< ... >'."
		}
		exitError $msg
	} else {
		return $token
	}
}


# getName --
#
#   Extract a token matching the pattern [A-Z].*
#
# return
#   the token

proc getName {} {
	set token [getToken]
	if {[regexp -- {^[A-Z]} $token]} {
		return $token
	} else {
		exitError "hsm name expected \na hsm name begin with a capital letter."
	}
}


# getBalanced --
#
#   Extract a text fragment between balanced <in> & <out>
#   The first <in> is considered as already eaten.
#
# return
#   the fragment

proc getBalanced {{in "\{"} {out "\}"}} {
	set level 1
	set fin false
	set res ""

	global parsed
	set parsedBefore $parsed

	while {1} {
		set token [getWord]
		if {$token == $out} {
			incr level -1
			if {$level == 0} { break }
		}
		if {$token == $in} { incr level }
		if {$token == "__EOF__"} {
			set parsed $parsedBefore
			exitError "EOF reached before end of block \na '$out' is probably missing."
		}
		append res $token
	}
	return $res
}


# getParentized --
#
#   Extract a word or a parentized expression
#
# return
#   the word or parentized expression

proc getParentized {} {
	set token [getToken]
	if {$token == "("} {
		set token "([getBalanced "(" ")"])"
	}
	return $token
}
