////////////////////////////////////////////////////////////////////////////////
// File    : Hsm.cpp
// Content : base classes for hierarchical state machines
// History : 26-Feb-2003 - [rb] creation
//           25-Mar-2003 - [rb] corrected a bug in SuperHsm::getHsm that
//                              returned always 0
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <Hsm.h>

#include <hsm/system/time/Timer.h>
#include <hsm/util/Functor.h>
#include <hsm/util/Debug.h>


namespace hsm {

/* Hsm class implementation ***************************************************/

	void Hsm::timeOut() {
		post(Event(this, Event::INTERNAL, 0));
	}
	
	Hsm::Hsm() : timerDelay(-1), timerFn(Functor::create(this, &Hsm::timeOut)), timer(timerDelay, timerFn), running(true) {}

	Hsm::~Hsm() {
		timer.stop();
		delete timerFn;
	}
	
	void Hsm::enable() {
		running = true;
	}
	
	void Hsm::disable() {
		running = false;
	}
	
	Hsm *Hsm::bottom() {
		return _current_->bottom();
	}
	
	Hsm *Hsm::top() {
		return _super_->top();
	}

	void Hsm::registerSubPath() {
		subPathMap[componentId] = this;
		for(Map::const_iterator h = hsmMap.begin(); h != hsmMap.end(); ++h) {
			(*h).second->registerSubPath();
		}
		for(Map::const_iterator h = subPathMap.begin(); h != subPathMap.end(); ++h) {
			_super_->subPathMap[(*h).first] = this;
		}
	}
	

	Hsm *Hsm::resolve(const Component::Id &id) {
		HSM_PRINT(1, "[" << componentId << "::Hsm::resolve(" << id << ")]")

		Hsm *hsm = subPathMap[id];
		
		// the target is not in my subHsms -> search upside
		if(hsm == 0) {
			doLeave(false);
			hsm = _super_->resolve(id);

		// the target is in my tree
		} else {
			_current_ = hsm;
			hsm = hsmMap[id];

			// the target is not my direct descendent -> searching inside
			if(hsm == 0) {
				if(_current_->doEnter(false, false) != 0) {
					return 0;
				}
				hsm = _current_->resolve(id);
			}
		}
		
		return hsm;
	}

	void Hsm::enterSubHsms(bool history /* = false */) {
		doEnter(history);
	}
	
	void Hsm::leaveSubHsms() {
		doLeave();
	}


	void Hsm::doInit() {
		HSM_PRINT(1, "[" << componentId << "(" << this << ")::Hsm::init]");
		_initial_ = 0;
		_init_();
	}

	void Hsm::doRegisterInputs() {
		registerInputs();
		
		for(Map::const_iterator h = hsmMap.begin(); h != hsmMap.end(); ++h) {
			((*h).second)->doRegisterInputs();
		};
	}


	Hsm *Hsm::doEnter(bool history /* = false */, bool rec /* = true */) {
		Hsm *except = checkRequis();
		if(except != 0) {
			HSM_PRINT(1, "REQUIREMENT FAILED in " <<  componentId << "::Hsm::enter")
			except->doEnter();
		} else {
			HSM_PRINT(1, "[" << componentId << "::Hsm::enter]");
			_enter_();
			timer.rearm(timerDelay);
			if(rec) {
				if(!history) _current_ = getInitial();
				_current_->doEnter(history);
			}
		}
		return except;
	}
			
	void Hsm::doLeave(bool rec /* = true */) {
		timer.stop();
		if(rec) {
			_current_->doLeave();
		}
		HSM_PRINT(1, "[" << componentId << "::Hsm::leave]");
		_leave_();	
	}
	
			
	void Hsm::doHandleEvent(const Event &event) {
		if(running && !handleEvent(event)) _current_->doHandleEvent(event);
	}
	
	
	void Hsm::broadcast(const Event &event) {
		dispatch(event);
	}
	
		
	void Hsm::process(const Event &event) {
		top()->doHandleEvent(event);
	}	
	
	
/* SubHsm class implementation ************************************************/

	SubHsm::SubHsm(Hsm *sup) {
		_super_ = sup;
	}
		
	Hsm *SubHsm::bottom() {
		return _super_;
	}
		
	Hsm *SubHsm::doEnter(bool history /* = false */, bool rec /* = true */) { return 0; } 
	void SubHsm::doLeave(bool rec /* = true */) {}
	
	void SubHsm::doHandleEvent(const Event &event) {
		HSM_PRINT(1, "Unhandled event : " << event)
	}
	

/* SuperHsm class implementation **********************************************/

	SuperHsm::SuperHsm(Hsm *cur) {
		hsmMap[cur->getId()] = _current_ = _initial_ = cur;
	}
		
	Hsm *SuperHsm::top() {
		return _current_;
	}


} // namespace hsm
