////////////////////////////////////////////////////////////////////////////////
// File    : Hsm.h
// Content : base classes for hierarchical state machines
// History : 26-Feb-2003 - [rb] creation
//           04-Feb-2004 - [rb] added the possibility to check assertions before
//                              entering a state
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// ToDo(-) / Done(+) ///////////////////////////////////////////////////////////
// - should Hsm::top & bottom be holded as variables ?
// - should event go from innermost to outermost or not ? (see doHandle and
//   process for implementation)
// - should Interface & Hsm virtually derivate from Component ?
// - should it be possible to choose a processing policy for Hsms ?


#if !defined(HSM_HSM_H)
#define HSM_HSM_H


/* includes *******************************************************************/

#include <hsm/core/Component.h>
#include <hsm/util/Template.h>
#include <hsm/system/time/Timer.h>

#include <map>


namespace hsm {

/* forward declarations *******************************************************/

	class Functor;


/* Interface class definition *************************************************/

	class Interface : public virtual impl::Component< > {};
	
	
/* Hsm class definition *******************************************************/

	class Hsm : public virtual impl::Component< > {
	// behaviours common to all states
	// members to overload :
	// - init           : add substates (set initial to the initial substate)
	// - checkRequis    : check if the state accept to be entered & returns
	//                    the state to enter
	// - enter, leave,
	//   escape         : add specific behaviour
	// - registerInputs : subscribe top() to the needed inputs for the state
	// - getInitial     : return the state to enter (should default to initial)
	// - handleEvent    : handle the event & return true if handled

		typedef std::map< Component::Id, Hsm * > Map;

	protected :
		Map subPathMap; ///< map of the path to the sub (and subsub ...) hsms
		
		Map hsmMap;     ///< map of the childs
		Hsm *_super_;   ///< way to the parent
		Hsm *_initial_; ///< initial state
		Hsm *_current_; ///< current state

		int      timerDelay; ///< time before fire (ms)
		Functor *timerFn;    ///< timer callback functor
		Timer    timer;      ///< timer object
		void     timeOut();  ///< timer callback
		
		bool running; ///< is the machine running ? (see Hsm::enable() and Hsm::disable())

	public :
		         Hsm(); ///< constructor
		virtual ~Hsm(); ///< destructor
		
		void enable();  ///< activate the machine
		void disable(); ///< desactivate the machine

	protected :
		virtual Hsm *bottom();  ///< pointer on the most inner active hsm
		virtual Hsm *top();     ///< pointer to the top machine
		
		virtual void registerSubPath(); ///< register the path to all the descendant of a hsm
		
	public :
		Hsm *resolve(const Component::Id &id);
		void enterSubHsms(bool history = false);
		void leaveSubHsms();

	protected :	
		void doInit(); ///< perform initialisation (mostly by calling Hsm::_init()) 
		void doRegisterInputs(); ///< register the hsm inputs
		
		virtual Hsm *doEnter(bool history = false, bool rec = true); 
		virtual void doLeave(bool rec = true);

		virtual void doHandleEvent(const Event &event); ///< handle an event (or transmit it to the current substate)

	public :
		void broadcast(const Event &event);

	protected :	
		virtual void process(const Event &event);

	protected :
		virtual void _init_() {}  ///< initialisation code (to be overridden)
		virtual void registerInputs() {}

		virtual void _enter_() {}  ///< enter code (to be overridden)
		virtual void _leave_() {}  ///< leave code (to be overridden)

		virtual bool handleEvent(const Event &event) { return false; }

		virtual Hsm *checkRequis() { return 0; }
		virtual Hsm *getInitial() { return _initial_; }
	};


/* SubHsm class definition ****************************************************/

	class SubHsm : public Hsm {
	// do nothing class to be used as substate for terminal states

	public :
		SubHsm(Hsm *sup);

	protected :
		virtual Hsm *bottom();

	protected :
		virtual Hsm *doEnter(bool history = false, bool rec = true); 
		virtual void doLeave(bool rec = true);

		virtual void doHandleEvent(const Event &event);
	};


/* SuperHsm class definition **************************************************/

	class SuperHsm : public Hsm {
	// do nothing class to be used as superstate for root state
	
	public :
		SuperHsm(Hsm *cur);

	protected :
		virtual Hsm *top();
	};


	namespace impl {
	
/* impl::Hsm template class definition ****************************************/

		template< const char *id, class H, class I, class Data > class Hsm : public H, public I {
		// classes derivated from AHsm have to be used in this template to run.
		// eg :
		//   struct MyData { ... };
		//   struct MyInterface : public MySuperInterface, protected virtual MyData { ... };
		//   class MyAHsm : public Hsm, protected virtual MyData { ... };
		//   extern const char name[];
		//   typedef impl::Hsm< name, MyAHsm, MyInterface, MyData > MyHsm;
			COMPILE_ASSERT(BASE_OF(hsm::Hsm, H), H_should_inherite_from_Hsm);
			COMPILE_ASSERT(BASE_OF(hsm::Interface, I), I_should_inherite_from_Interface);

		public :
			Data &data;

		protected :
			void construct(hsm::Hsm *sup = 0) {
				componentId = id;
				_super_ = sup;

				doInit();
				if(_initial_ == 0) _initial_ = new SubHsm(this);
				if(_super_ == 0) {
					_super_ = new SuperHsm(this);
					registerSubPath();
					doEnter();
					doRegisterInputs();
				}
			}

		public :
			Hsm< id, H, I, Data >(hsm::Hsm *sup, Data d) : Data(d), data(*this) { construct(sup); }
			Hsm< id, H, I, Data >(hsm::Hsm *sup = 0) : Data(), data(*this)      { construct(sup); }
			Hsm< id, H, I, Data >(Data d) : Data(d), data(*this)                { construct(); }

			virtual ~Hsm< id, H, I, Data >() { if(top() == this) doLeave(); }

			I *operator->() { return dynamic_cast< I * >(bottom()); }
		};
		
		
	} //namespace impl

} // namespace hsm


#endif // !defined(HSM_HSM_H)