////////////////////////////////////////////////////////////////////////////////
// File    : Component.cpp
// Content : a common interface class for hsm component
// History : 16-Oct-2001 - [rb] creation
//           03-Mar-2003 - [rb] change receiver list to set (permit multiple
//                              subscriptions).
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <Component.h>

#include <hsm/core/Composite.h>
#include <hsm/system/window/EventHandler.h>
#include <hsm/system/plugin/PlugIn.h>

#include <sstream>


namespace hsm {
	
/* Component class implementation **************************************************/

	unsigned int Component::cn = 0;
	
	Component::Component(const Id &id /* = Id() */) {
		if(id != Id()) {
			componentId = id;
		} else {
			std::stringstream name;
			name << "component_" << (cn++);
			componentId = name.str();
		}
	}

	Component::~Component() {
		dispatch(Event(this, Event::DELETED));
		for(ComponentSet::iterator c, e = emitters.begin(); e != emitters.end();) {
			c = e; ++e;
			(*c)->removeReceiver(this);
		}
	}
	
	Component::Id Component::getId() const {
		return componentId;
	}

	// emitter ------------------------------------
	void Component::addReceiver(Component *receiver) {
		receivers.insert(receiver);
		receiver->addEmitter(this);
	}

	void Component::removeReceiver(Component *receiver) {
		receiver->removeEmitter(this);
		receivers.erase(receiver);
	}

	void Component::dispatch(const Event &event) {
		for(ComponentSet::iterator r = receivers.begin(); r != receivers.end(); ++r) {
			(*r)->post(event);
		}
	}
	
	void Component::dispatch(Event::Message m /* = Event::CHANGE */, Data d /* = 0 */) {
		dispatch(Event(this, m, d));
	}

	
	// receiver -----------------------------------
	void Component::post(const Event &event) {
		if(event.message == Event::DELETED) {
			emitters.erase(event.sender);
		}
		doPost(event);
	}
	
	void Component::post(Event::Message m /* = Event::CHANGE */, Data d /* = 0 */) {
		post(Event(0, m, d));
	}

	void Component::addEmitter(Component *emitter) {
		emitters.insert(emitter);
	}
	
	void Component::removeEmitter(Component *emitter) {
		emitters.erase(emitter);
	}


	namespace repository {
		
/* static data initialisation *************************************************/
	
		static impl::Component< Composite > Root("Plugin Root");
	
	
/* functions ******************************************************************/
	
		Component *getComponent(const Component::Id &id) {
			Component *res = 0;
			if(id == "Plugin Root") {
				res = &Root;
			} else {
				res = Root.get(id);
				if(res == 0)	{
					plugin::load("components/" + id);
					res = Root.get(id);
				}
			}
			return res;
		}
	
		EventHandler *getEventHandler(const Component::Id &id) {
			return dynamic_cast< EventHandler * >(getComponent(id));
		}
	
		void add(Component *component) {
			Root.add(component);
		}
	
		void remove(Component *component) {
			Root.remove(component);
		}
		

	} // namespace repository
	
} // namespace hsm
