////////////////////////////////////////////////////////////////////////////////
// File    : Window.MACOSX.cpp
// Content : Window::os struct implementation
// History : 11-Oct-2001 - [rb] creation
//           01-Feb-2004 - [rb] put the window to foreground
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <Window.h>

#include <hsm/component/Point.h>
#include <hsm/system/window/EventHandler.h>

#include <Carbon/Carbon.h>


namespace hsm {

/* static variable initialisation *********************************************/

	Window::os::Map Window::os::windowMap;


/* Window::os struct implementation *******************************************/

	OSStatus Window::os::defaultHandler(EventHandlerCallRef inRef, EventRef inEvent, void *userData) {
		WindowRef window;
		Rect rect;

		switch(GetEventClass(inEvent)){
		case kEventClassWindow :
			GetEventParameter(inEvent, kEventParamDirectObject, typeWindowRef, 0, sizeof(WindowRef), 0, &window);
			hsm::Window *hw = windowMap[window];

			switch(GetEventKind(inEvent)){
			case kEventWindowClosed :
				exit(EXIT_SUCCESS);
				break;

			case kEventWindowBoundsChanged :
				GetEventParameter(inEvent, kEventParamCurrentBounds, typeQDRectangle, 0, sizeof(Rect), 0, &rect);
				*(hw->position) = rect.left, rect.top;
				*(hw->size) = (rect.right - rect.left), (rect.bottom - rect.top);
				break;
			}
		}
		
		return eventNotHandledErr;
	}

			
	Window::os::Window Window::os::create(hsm::Window *hw, const std::string &title, int x, int y, int w, int h) {
		Rect rect;
		SetRect(&rect, x, y, x + w, y + h);
		
		WindowRef window;
		CreateNewWindow(kDocumentWindowClass,
		                kWindowStandardDocumentAttributes | kWindowStandardHandlerAttribute,
		                &rect, &window);
		Str255 sTitle;
		CopyCStringToPascal(title.c_str(), sTitle);
		SetWTitle(window, sTitle);
		
		windowMap[window] = hw;
		setHandler(window);
		
		return window;
	}
	
	void Window::os::destroy(Window window) {
		ReleaseWindow(window);
	}

	void Window::os::setHandler(Window window) {
		EventTypeSpec events[] = { {kEventClassWindow, kEventWindowClosed},
		                           {kEventClassWindow, kEventWindowBoundsChanged} };
		InstallWindowEventHandler(window, NewEventHandlerUPP(defaultHandler),
		                          GetEventTypeCount(events), events,
		                          0, 0);	
	}
	
	void Window::os::setPosition(Window window, int x, int y, int w, int h) {
		MoveWindow(window, x, y, true);
	}
	
	void Window::os::setSize(Window window, int x, int y, int w, int h) {
		SizeWindow(window, w, h, false);
	}
	
	void Window::os::setTitle(Window window, const std::string &title) {
		Str255 sTitle;
		CopyCStringToPascal(title.c_str(), sTitle);
		SetWTitle(window, sTitle);	
	}
			
	void Window::os::show(Window window) {
		ShowWindow(window);
	}
	
	void Window::os::hide(Window window) {
		HideWindow(window);
	}

	void Window::os::zoom(Window window) {
		ZoomWindow(window, inZoomOut, true);
	}

	void Window::os::unzoom(Window window) {
		ZoomWindow(window, inZoomIn, true);
	}

			
	void Window::os::showCursor() {
		CGDisplayShowCursor(kCGDirectMainDisplay);
	}

	void Window::os::hideCursor() {
		CGDisplayHideCursor(kCGDirectMainDisplay);
	}

	void Window::os::addHandler(hsm::Window *hw, const EventHandler *eh) {
		EventHandlerRef handlerId;
		InstallWindowEventHandler(hw->window, NewEventHandlerUPP(eh->getHandler()),
		                          eh->nbEvents, eh->eventSpec,
		                          0,
		                          &handlerId);
		hw->handlers[eh] = handlerId;
	}
	
	void Window::os::removeHandler(hsm::Window *hw, const EventHandler *eh) {
		RemoveEventHandler(hw->handlers[eh]);
	}


/* loading action *************************************************************/

	namespace {
		
		class Loader {
		private :
			static const Loader instance;
			
		private :
			Loader() {
				ProcessSerialNumber psn;
				GetCurrentProcess(&psn);
				SetFrontProcess(&psn);
			}

		};

		const Loader Loader::instance;

	} // namespace

} // namespace hsm
