////////////////////////////////////////////////////////////////////////////////
// File    : Debug.h
// Content : debug macros
// History : 05-Aou-2001 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Debug.h
/// Usefull debug functions & macros.


#if !defined(HSM_DEBUG_H)
#define HSM_DEBUG_H


/* includes *******************************************************************/

#include <cstdlib>
#include <iostream>


namespace hsm {

	/// contains functions related to debugging
	namespace debug {

/* declarations ***************************************************************/

		/// Gets the debug level.
		/// The debug level is used by HSM_PRINT() to choose the debugging
		/// information to display.
		/// \see setLevel()
		/// \return the debug level
		unsigned int getLevel();

		/// Sets the debug level.
		/// \see getLevel()
		/// \param d the debug level
		void setLevel(unsigned int d);


	} // namespace debug

} // namespace hsm


/* macros *********************************************************************/

/// \def HSM_DEBUG(i)
/// Expands to \a i if DEBUG_VERSION is set else to nop.
/// \param i instruction

/// \def HSM_FINAL(i)
/// Expands to nop if DEBUG_VERSION is set else to \a i.
/// \param i instruction

/// \def HSM_PRINT(l, s)
/// Prints \a s on \a cout if \a l <= getDebugLevel() and if DEBUG_VERSION is
/// set else do nothing.
/// \param l debug level
/// \param s message

#if defined(DEBUG_VERSION)
#  define HSM_DEBUG(i) i
#  define HSM_FINAL(i)
#  define HSM_PRINT(l, s) { \
		if(l <= hsm::debug::getLevel()) { std::cout << s << std::endl; } \
	}
#else
#  define HSM_DEBUG(i)
#  define HSM_FINAL(i) i
#  define HSM_PRINT(l, s)
#endif


/// Prints the file and line information and the \a s warning message on
/// \c std::cerr.
/// \param s message
#define HSM_WARNING(s) { \
	std::cerr << __FILE__ << "(" << __LINE__ << ") WARNING : " << s << std::endl; \
}

/// Prints the file and line information and the \a s error message on
/// \c std::cerr and then exits with status EXIT_FAILURE.
/// \param s message
#define HSM_FAILURE(s) { \
	std::cerr << __FILE__ << "(" << __LINE__ << ") FAILURE : " << s << std::endl;  \
	exit(EXIT_FAILURE); \
}


#define HSM_OUTPUT(s)  { std::cout << s << std::endl; } ///< Prints \a s on \c std::cout.
#define HSM_ERRPUT(s)  { std::cerr << s << std::endl; } ///< Prints \a s on \c std::cerr.


#endif // !defined(HSM_DEBUG_H)
