////////////////////////////////////////////////////////////////////////////////
// File    : Functor.h
// Content : class prototype for a function object
// History : 09-Jan-2002 - [rb] creation
//           18-Jan-2002 - [rb] added support for void (*pf)(void *) functions
//                         and makers for non template subclasses
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Functor.h
/// Class and helpers to manipulate functions as objects.


#if !defined(HSM_FUNCTOR_H)
#define HSM_FUNCTOR_H


/* includes *******************************************************************/

#include <map>
#include <string>


namespace hsm {

/* Functor class defininition *************************************************/

	/// Implements a default do-nothing function operator.
	class Functor {
		typedef std::map< std::string, Functor * > FunctorMap;
		static FunctorMap functors;
		
	public :
		/// function operator with optionnal \c void pointer parameter doing nothing
		/// \param pData optionnal pointer on user data
		virtual void operator()(void *pData = 0) {}

		/// \{
		/// polymorphicaly creates a functor
		static Functor *create(void (*pf)());
		static Functor *create(void (*pf)(void *));
		template< class C > static Functor *create(C *pc, void (C::*pf)());
		template< class C > static Functor *create(C *pc, void (C::*pf)(void *));
		/// \}
		
		static void add(const std::string &id, Functor *fn);
		static void erase(const std::string &id);
		static Functor *get(const std::string &id);
		
		static void call(const std::string &id, void *arg = 0);
	};


/* FunctionFunctor class definition *******************************************/

	/// Implements a functor encapsulating a void fn(void).
	class FunctionFunctor : public Functor {
	private :
		void (*ptrFn)();

	public :

		/// constructor
		/// \param pf pointer on a \a void \a f() function
		FunctionFunctor(void (*pf)()) : ptrFn(pf) {}

		/// do it function operator
		/// \param pData optionnal pointer on user data (not passed to the actual function)
		virtual void operator()(void *pData = 0) { (*ptrFn)(); }
	};


/* FunctionPVoidFunctor class definition **************************************/

	/// Implements a functor encapsulating a void fn(void *).
	class FunctionPVoidFunctor : public Functor {
	private :
		void (*ptrFn)(void *);

	public :

		/// constructor
		/// \param pf pointer on a \a void \a f(void *) function
		FunctionPVoidFunctor(void (*pf)(void *)) : ptrFn(pf) {}

		/// do it function operator
		/// \param pData optionnal pointer on user data (passed to the actual function)
		virtual void operator()(void *pData = 0) { (*ptrFn)(pData); }
	};


/* ObjectFunctor template class definition ************************************/

	/// Implements a functor encapsulating an object member
	template< class C > class ObjectFunctor : public Functor {
	private :
		C *ptrC;
		void (C::*ptrFn)();
		
	public :
	
		/// constructor
		/// \param pc pointer on an object instance
		/// \param pf pointer on a \a void \a f() object member
		ObjectFunctor(C *pc, void (C::*pf)()) : ptrC(pc), ptrFn(pf) {}

		/// do it function operator
		/// \param pData optionnal pointer on user data (not passed to the actual function)
		virtual void operator()(void *pData = 0) { (ptrC->*ptrFn)(); }
	};


/* ObjectPVoidFunctor template class definition *******************************/

	/// Implements a functor encapsulating an object member
	template< class C > class ObjectPVoidFunctor : public Functor {
	private :
		C *ptrC;
		void (C::*ptrFn)(void *);
		
	public :
	
		/// constructor
		/// \param pc pointer on an object instance
		/// \param pf pointer on a \a void \a f() object member
		ObjectPVoidFunctor(C *pc, void (C::*pf)(void *)) : ptrC(pc), ptrFn(pf) {}

		/// do it function operator
		/// \param pData optionnal pointer on user data (not passed to the actual function)
		virtual void operator()(void *pData = 0) { (ptrC->*ptrFn)(pData); }
	};


/* Functor template function definition ***************************************/

	template< class C > Functor *Functor::create(C *pc, void (C::*pf)()) {
		return new ObjectFunctor< C >(pc, pf);
	}

	template< class C > Functor *Functor::create(C *pc, void (C::*pf)(void *)) {
		return new ObjectPVoidFunctor< C >(pc, pf);
	}


} // namespace hsm


#endif // !defined(HSM_FUNCTOR_H)
